/**
 * @author: @AngularClass
 */

const webpack = require('webpack');
const helpers = require('./helpers');
const ngtools = require('@ngtools/webpack');
const branding = require('./branding');
const path = require('path');
const stringify = require('json-stringify');
const fs = require('fs');

/**
 * Webpack Plugins
 */
const autoprefixer = require('autoprefixer');
//const CheckerPlugin = require('awesome-typescript-loader').CheckerPlugin;
const ContextReplacementPlugin = require('webpack/lib/ContextReplacementPlugin');
const CopyWebpackPlugin = require('copy-webpack-plugin');
const DefinePlugin = require('webpack/lib/DefinePlugin');
const ExtractTextPlugin = require('extract-text-webpack-plugin');
const HtmlWebpackPlugin = require('html-webpack-plugin');
const IgnorePlugin = require('webpack/lib/IgnorePlugin');
const LoaderOptionsPlugin = require('webpack/lib/LoaderOptionsPlugin');
// const ngcWebpack = require('ngc-webpack');
const MiniCssExtractPlugin = require('mini-css-extract-plugin');
const NormalModuleReplacementPlugin = require('webpack/lib/NormalModuleReplacementPlugin');
const ProvidePlugin = require('webpack/lib/ProvidePlugin');
const ScriptExtHtmlWebpackPlugin = require('script-ext-html-webpack-plugin');
const StyleLintPlugin = require('stylelint-webpack-plugin');
const ManifestPlugin = require('webpack-manifest-plugin');
const ForkTsCheckerWebpackPlugin = require('fork-ts-checker-webpack-plugin');
const IgnoreNotFoundExportPlugin = require('./IgnoreNotFoundExportPlugin');

/**
 * Webpack Constants
 */
// const HMR = helpers.hasProcessFlag('hot');
// const AOT = helpers.hasNpmFlag('aot');
const METADATA = {
  baseUrl: '/',
  isDevServer: helpers.isWebpackDevServer(),
  FABRIC8_BRANDING: process.env.FABRIC8_BRANDING || 'fabric8',
  PUBLIC_PATH: process.env.PUBLIC_PATH || '/',
};

/*
 * Webpack configuration
 *
 * See: http://webpack.github.io/docs/configuration.html#cli
 */
module.exports = function (options) {
  const isProd = options.env === 'production';
  const aotMode = false;//options && options.aot !== undefined;
  console.log('The options from the webpack config: ' + stringify(options, null, 2));

  // const entryFile = aotMode ? './src/main.browser.aot.ts' : './src/main.browser.ts';
  // const outPath = aotMode ? 'dist' : 'aot';
  // const devtool = aotMode ? 'source-map' : 'eval-source-map';
  // const srcPath = path.join(__dirname, 'demo', 'src');
  var config = {

    /*
     * Cache generated modules and chunks to improve performance for multiple incremental builds.
     * This is enabled by default in watch mode.
     * You can pass false to disable it.
     *
     * See: http://webpack.github.io/docs/configuration.html#cache
     */
    //cache: false,

    /*
     * The entry point for the bundle
     * Our Angular.js app
     *
     * See: https://webpack.js.org/configuration/entry-context/#entry
     */
    entry: {
      'vendor': './src/vendor.browser.ts',
      'polyfills': './src/polyfills.browser.ts',
      // 'main': aotMode ? './src/main.browser.aot.ts' : './src/main.browser.ts'
      'main': [
        './src/main.browser.ts',
        // workaround for https://github.com/webpack-contrib/extract-text-webpack-plugin/issues/456
        // 'style-loader/lib/addStyles',
        // 'css-loader/lib/css-base',
      ]
    },

    /**
     * Options affecting the resolving of modules.
     *
     * See: https://webpack.js.org/configuration/resolve
     */
    resolve: {

      /**
       * An array that automatically resolve certain extensions.
       * Which is what enables users to leave off the extension when importing.
       *
       * See: https://webpack.js.org/configuration/resolve/#resolve-extensions
       */
      extensions: ['.ts', '.js', '.json']
    },

    /*
     * Options affecting the normal modules.
     *
     * See: http://webpack.github.io/docs/configuration.html#module
     */
    module: {

      rules: [

        /*
         * Typescript loader support for .ts and Angular 2 async routes via .async.ts
         * Replace templateUrl and stylesUrl with require()
         *
         * See: https://github.com/s-panferov/awesome-typescript-loader
         * See: https://github.com/TheLarkInn/angular2-template-loader
         */
        {
          test: /\.ts$/,
          use: aotMode ? [
            '@ngtools/webpack'
          ] : [
              'cache-loader',
              {
                loader: 'thread-loader',
                options: {
                    // there should be 1 cpu for the fork-ts-checker-webpack-plugin
                    workers: require('os').cpus().length - 1,
                },
              },
              {
                loader: 'ts-loader',
                options: {
                    happyPackMode: true
                }
              },
              'angular2-template-loader',
              'angular2-router-loader'
            ],
          // loaders: '@ngtools/webpack',
          exclude: [/\.(spec|e2e)\.ts$/]
        },

        /**
         * Static analysis linter for TypeScript advanced options configuration
         * Description: An extensible linter for the TypeScript language.
         *
         * See: https://github.com/wbuchwalter/tslint-loader
         */
        {
          test: /\.ts$/,
          enforce: 'pre',
          loader: 'tslint-loader',
          exclude: [
            /node_modules/,
            /src\/a-runtime-console/,
            // /src\/main.browser.ts/,
            // /src\/polyfills.browser.ts/,
            // /src\/vendor.browser.ts/
          ],
          options: {
          //   configuration: {
          //     rules: {
          //       quotemark: [true, 'double']
          //     }
          //   },
          //
            // can specify a custom config file relative to current directory or with absolute path
            // 'tslint-custom.json'
            configFile: "tslint.json",

            // tslint errors are displayed by default as warnings
            // set emitErrors to true to display them as errors
            emitErrors: true,

            // tslint does not interrupt the compilation by default
            // if you want any file with tslint errors to fail
            // set failOnHint to true
            failOnHint: true,

            // enables type checked rules like 'for-in-array'
            // uses tsconfig.json from current working directory
            typeCheck: false,

            // automatically fix linting errors
            fix: false,

            // can specify a custom tsconfig file relative to current directory or with absolute path
            // to be used with type checked rules
            tsConfigFile: 'tsconfig.json',

            // path to directory containing formatter (optional)
            formattersDirectory: 'node_modules/tslint-loader/formatters/',

            // name of your formatter (optional)
            formatter: 'custom',

            // These options are useful if you want to save output to files
            // for your continuous integration server
            fileOutput: {
              // The directory where each file's report is saved
              dir: './lint/',

              // The extension to use for each report's filename. Defaults to 'txt'
              ext: 'xml',

              // If true, all files are removed from the report directory at the beginning of run
              clean: true,

              // A string to include at the top of every report file.
              // Useful for some report formats.
              header: '<?xml version="1.0" encoding="utf-8"?>\n<checkstyle version="5.7">',

              // A string to include at the bottom of every report file.
              // Useful for some report formats.
              footer: '</checkstyle>'
            }
          }
        },

        /**
         * Json loader support for *.json files.
         *
         * See: https://github.com/webpack/json-loader
         */
        {
          test: /\.json$/,
          type: "javascript/auto",
          use: ['json-loader']
        },

        /**
         * HTML Linter
         * Checks all files against .htmlhintrc
         */
        {
          enforce: 'pre',
          test: /\.html$/,
          use: {
            loader: 'htmlhint-loader',
            options: {
              configFile: './.htmlhintrc'
            }
          },
          exclude: [/node_modules/, /src\/a-runtime-console/]
        },

        /**
         * Raw loader support for *.html
         * Returns file content as string
         *
         * See: https://github.com/webpack/raw-loader
         */
        {
          test: /\.html$/,
          use: ['html-loader']
        },

        /**
         * to string and css loader support for *.css files
         * Returns file content as string
         *
         */
        {
          test: /^(?!.*component).*\.css$/,
          use: [
            isProd ? MiniCssExtractPlugin.loader : 'style-loader',
            {
              loader: 'css-loader',
              options: {
                sourceMap: !isProd,
                context: '/'
              },
            },
          ]
        },
        {
          test: /\.component\.css$/,
          use: [
            'to-string-loader',
            {
              loader: 'css-loader',
              options: {
                minimize: true,
                sourceMap: !isProd,
                context: '/'
              }
            }
          ]
        },

        {
          test: /^(?!.*component).*\.less$/,
          use: [
            isProd ? MiniCssExtractPlugin.loader : 'style-loader',
            {
              loader: 'css-loader',
              options: {
                minimize: true,
                sourceMap: !isProd,
                context: '/'
              }
            }, {
              loader: 'less-loader',
              options: {
                paths: [
                  path.resolve(__dirname, "../node_modules"),
                  path.resolve(__dirname, "../node_modules/patternfly/dist/less"),
                  path.resolve(__dirname, "../node_modules/patternfly/dist/less/dependencies"),
                  path.resolve(__dirname, "../node_modules/patternfly/dist/less/dependencies/bootstrap"),
                  path.resolve(__dirname, "../node_modules/patternfly/dist/less/dependencies/font-awesome"),
                ],
                sourceMap: !isProd
              }
            }
          ]
        }, {
          test: /\.component\.less$/,
          use: [
            'to-string-loader',
            {
              loader: 'css-loader',
              options: {
                minimize: true,
                sourceMap: !isProd,
                context: '/'
              }
            }, {
              loader: 'less-loader',
              options: {
                paths: [
                  path.resolve(__dirname, "../node_modules"),
                  path.resolve(__dirname, "../node_modules/patternfly/dist/less"),
                  path.resolve(__dirname, "../node_modules/patternfly/dist/less/dependencies"),
                  path.resolve(__dirname, "../node_modules/patternfly/dist/less/dependencies/bootstrap"),
                  path.resolve(__dirname, "../node_modules/patternfly/dist/less/dependencies/font-awesome"),
                ],
                sourceMap: !isProd
              }
            }
          ]
        },

        /**
         * File loader for supporting fonts, for example, in CSS files.
         */
        {
          test: /\.(woff2|woff|ttf|eot|svg)$/,
          use: {
            loader: 'url-loader',
            options: {
              limit: 3000,
              includePaths: [
                path.resolve(__dirname, "../node_modules/patternfly/dist/fonts/")
              ],
              name: '_assets/fonts/[name]' + (isProd ? '.[hash]' : '') + '.[ext]'
            }
          },
          exclude: [
            path.resolve(__dirname, "../src/assets/images/"),
            /OpenSans.*\.(woff2|woff|ttf|eot|svg)$/ // Exclude loading OpenSans fonts as it will come from CDN
          ]
        },
        {
          test: /\.(jpg|png|svg|gif|jpeg)$/,
          use: {
            loader: 'url-loader',
            options: {
              limit: 3000,
              includePaths: [
                path.resolve(__dirname, "../src/assets/images/")
              ],
              name: '_assets/images/[name]' + (isProd ? '.[hash]' : '') + '.[ext]'
            }
          },
          exclude: path.resolve(__dirname, "../node_modules/patternfly/dist/fonts/")
        }
      ]
    },

    /**
     * Add additional plugins to the compiler.
     *
     * See: http://webpack.github.io/docs/configuration.html#plugins
     */
    plugins: [
      /**
       * Plugin: ForkTsCheckerWebpackPlugin
       * Description: Webpack plugin that runs typescript type checker on a separate process.
       *
       * See: https://github.com/Realytics/fork-ts-checker-webpack-plugin
       */
      new ForkTsCheckerWebpackPlugin({
        // Must set to `true` if using ts-loader `happyPackMode === true`
        checkSyntacticErrors: true,
        // We want to block webpack's emit and wait for the checker to complete and add any errors
        // to webpack's compilation.
        async: false
      }),

      new IgnoreNotFoundExportPlugin(),

      /**
       * Plugin: ContextReplacementPlugin
       * Description: Provides context to Angular's use of System.import
       *
       * See: https://webpack.github.io/docs/list-of-plugins.html#contextreplacementplugin
       * See: https://github.com/angular/angular/issues/11580
       */
      new ContextReplacementPlugin(
        // The (\\|\/) piece accounts for path separators in *nix and Windows
        // /angular(\\|\/)core(\\|\/)@angular/,
        /\@angular(\\|\/)core(\\|\/)fesm5/,
        helpers.root('./src')
      ),
      new ContextReplacementPlugin(
        // The (\\|\/) piece accounts for path separators in *nix and Windows
        /angular(\\|\/)core(\\|\/)(esm(\\|\/)src|src)(\\|\/)linker/,
        helpers.root('src') // location of your src
      ),

      /*
       * Plugin: CopyWebpackPlugin
       * Description: Copy files and directories in webpack.
       *
       * Copies project static assets.
       *
       * See: https://www.npmjs.com/package/copy-webpack-plugin
       */
       /*
       * this needs to be redirected to reside in the _assets/ directory
       */
      new CopyWebpackPlugin([
        {
          from: 'src/meta'
        }
      ]),

      /*
       * Plugin: HtmlWebpackPlugin
       * Description: Simplifies creation of HTML files to serve your webpack bundles.
       * This is especially useful for webpack bundles that include a hash in the filename
       * which changes every compilation.
       *
       * See: https://github.com/ampedandwired/html-webpack-plugin
       */
      new HtmlWebpackPlugin({
        template: 'src/index.ejs',
        title: branding.assets[METADATA.FABRIC8_BRANDING].title.prefix,
        metadata: METADATA
      }),

      /*
       * Plugin: ScriptExtHtmlWebpackPlugin
       * Description: Enhances html-webpack-plugin functionality
       * with different deployment options for your scripts including:
       *
       * See: https://github.com/numical/script-ext-html-webpack-plugin
       */
      new ScriptExtHtmlWebpackPlugin({
        defaultAttribute: 'defer'
      }),

      /*
       * Plugin LoaderOptionsPlugin (experimental)
       *
       * See: https://gist.github.com/sokra/27b24881210b56bbaff7
       */
      new LoaderOptionsPlugin({}),

      // ExtractTextPlugin
      // const extractCSS = new ExtractTextPlugin({
      //   filename: '_assets/stylesheets/[name].[id]' + ( isProd ? '.[contenthash]' : '' ) + '.css'
      // });

      // MiniCssExtractPlugin
      new MiniCssExtractPlugin({
        // filename: '[name].css',
        filename: '_assets/stylesheets/[name]' + ( isProd ? '.[contenthash:8]' : '' ) + '.css',
        chunkFilename: '_assets/stylesheets/[id]' + ( isProd ? '.[contenthash:8]' : '' ) + '.css'
      }),

      new webpack.IgnorePlugin(/^\.\/locale$/, /[^-]moment$/),

      // Generate a manifest file which contains a mapping of all asset filenames
      // to their corresponding output file so that tools can pick it up without
      // having to parse `index.html`.
      new ManifestPlugin({
        fileName: 'asset-manifest.json',
        publicPath: METADATA.PUBLIC_PATH,
      }),

      /**
       * StyleLintPlugin
       */
      new StyleLintPlugin({
        configFile: '.stylelintrc',
        syntax: 'less',
        context: 'src',
        files: '**/*.less',
        lintDirtyModulesOnly: false,
        failOnError: false,
        emitErrors: true,
        quiet: false,
      })
    ],

     /**
     * Version 4 webpack runs optimizations for you depending on the chosen mode.
     *
     * The following plugins have been removed from Webpack 4 which were extensively used in previous versions.
     *  - NoEmitOnErrorsPlugin
     *  - ModuleConcatenationPlugin
     *  - NamedModulesPlugin
     *  - CommonsChunkPlugin
     *
     * see: https://webpack.js.org/configuration/optimization/
     */
    optimization: {
      namedModules: true, // NamedModulesPlugin()
      // Keep the runtime chunk seperated to enable long term caching
      // https://twitter.com/wSokra/status/969679223278505985
      runtimeChunk: true,
      splitChunks: { // CommonsChunkPlugin()
        chunks: 'async'
      },
      noEmitOnErrors: true, // NoEmitOnErrorsPlugin
      concatenateModules: true //ModuleConcatenationPlugin
    },

    /**
     * Include polyfills or mocks for various node stuff
     * Description: Node configuration
     *
     * See: https://webpack.github.io/docs/configuration.html#node
     */
    node: {
      clearImmediate: false,
      crypto: 'empty',
      global: true,
      module: false,
      process: true,
      setImmediate: false
    }
  };

  if (aotMode) {
    config.plugins.push(new ngtools.AotPlugin({
      tsConfigPath: 'tsconfig-aot.json'
      // entryModule: './src/app/app.module#AppModule',
      // genDir: './src/aot'
    }));
  }

  return config;
};
