import { Params } from '@angular/router';
import { currentOAuthConfig } from '../store/oauth-config-store';
import { AppInfo, Build, Builds, ServiceEnvironments, ServiceUrl } from './build.model';
import { KubernetesSpecResource } from './kuberentesspecresource.model';
import { pathJoin } from './utils';

export const defaultBuildIconStyle = 'pficon-build';

export class BuildConfig extends KubernetesSpecResource {
  gitUrl: string;

  type: string;

  lastVersion: number;

  jenkinsJobUrl: string;

  editPipelineUrl: string;

  openInCheUrl: string;

  lastBuildPath: string;

  lastBuildName: string;

  // last build related data
  statusPhase: string;

  duration: number;

  iconStyle: string;

  private _builds: Array<Build> = [];

  interestingBuilds: Array<Build> = [];

  private _lastBuild: Build;

  get serviceUrls(): Array<ServiceUrl> {
    const last = this.lastBuild;
    return last ? last.serviceUrls : [];
  }

  get builds(): Array<Build> {
    return this._builds;
  }

  set builds(builds: Array<Build>) {
    if (builds) {
      builds.sort((a: Build, b: Build) => b.buildNumberInt - a.buildNumberInt);
    }
    this._builds = builds;
    this.onBuildsUpdated();
  }

  get lastBuild(): Build {
    return this._lastBuild;
  }

  /**
   * When the builds are updated lets keep track of the interesting builds
   * and the last build
   */
  private onBuildsUpdated() {
    const answer = this.interestingBuilds;
    answer.splice(0, answer.length);

    const builds = this.builds;
    const lastVersionName = this.lastBuildName;
    if (lastVersionName) {
      for (const build of builds) {
        if (lastVersionName === build.name) {
          this._lastBuild = build;
        }
      }
    }
    if (!this._lastBuild && builds.length) {
      this._lastBuild = builds[0];
      if (!this.lastBuildName && this._lastBuild) {
        this.lastBuildName = this._lastBuild.name;
      }
    }

    for (const build of builds) {
      if (build.statusPhase === 'Running') {
        answer.push(build);
      }
    }
    const build = this.lastBuild;
    if (!answer.length && build) {
      answer.push(build);
    }

    this.statusPhase = build ? build.statusPhase : '';
    this.duration = build ? build.duration : 0;
    this.iconStyle = build ? build.iconStyle : defaultBuildIconStyle;
  }

  get isPipeline(): boolean {
    return this.type === 'JenkinsPipeline';
  }

  get interestingBuildsAverageDuration(): number {
    let answer = 0;
    let count = 0;
    const builds = this.interestingBuilds;
    for (const build of builds) {
      const duration = build.duration;
      if (duration) {
        answer += duration;
        count++;
      }
    }
    if (count > 0) {
      return answer / count;
    }
    return 0;
  }

  /**
   * Returns the Jenkins test report URL of the last build if it is available
   */
  get jenkinsTestReportUrl(): string {
    const build = this.lastBuild;
    return build ? build.jenkinsTestReportUrl : '';
  }

  get serviceEnvironmentMap(): Map<string, ServiceEnvironments> {
    const build = this.lastBuild;
    if (!build) {
      return new Map<string, ServiceEnvironments>();
    }
    const answer = build.serviceEnvironmentMap;
    const builds = this.builds;
    if (builds.length && builds.length > 1) {
      const previousBuild = builds[1];
      const map = previousBuild.serviceEnvironmentMap;
      if (map) {
        for (const key in map) {
          if (key && map[key]) {
            const value = map[key];
            if (!answer[key]) {
              answer[key] = value;
            }
          }
        }
      }
    }
    return answer;
  }

  /**
   * Returns a map indexed by the environment key of the app information
   */
  get environmentApp(): Map<string, AppInfo> {
    const map = this.serviceEnvironmentMap;
    const answer = new Map<string, AppInfo>();
    const name = this.name;
    if (map && name) {
      for (const environmentKey in map) {
        if (environmentKey && map[environmentKey]) {
          const value = map[environmentKey];
          const appInfo = value.toAppInfo(name);
          if (appInfo) {
            answer[environmentKey] = appInfo;
          }
        }
      }
    }
    return answer;
  }

  updateValuesFromResource() {
    super.updateValuesFromResource();

    const spec = this.spec || {};
    const status = this.status || {};
    const source = spec.source || {};
    const git = source.git || {};
    const strategy = spec.strategy || {};
    const type = strategy.type || '';

    this.lastVersion = status.lastVersion || 0;
    this.lastBuildName = this.lastVersion ? `${this.name}-${this.lastVersion}` : '';
    this.lastBuildPath = this.lastBuildName ? `${this.name}/builds/${this.lastBuildName}` : '';
    this.iconStyle = defaultBuildIconStyle;

    this.type = type;
    let gitUrl = this.annotations['fabric8.io/git-clone-url'];
    if (!gitUrl) {
      gitUrl = git.uri || '';
    }
    this.gitUrl = gitUrl;
    this.jenkinsJobUrl = this.annotations['fabric8.link.jenkins.job/url'] || '';
    this.onBuildsUpdated();

    const name = this.name;
    const namespace = this.namespace;
    const oauthConfig = currentOAuthConfig();
    if (oauthConfig) {
      const openShiftConsoleUrl = oauthConfig.openshiftConsoleUrl;
      if (openShiftConsoleUrl && namespace && name) {
        this.editPipelineUrl = pathJoin(
          openShiftConsoleUrl,
          '/project/',
          namespace,
          '/edit/pipelines/',
          name,
        );
      }
    }

    // TODO create openInCheUrl URL
  }

  defaultKind() {
    return 'BuildConfig';
  }

  defaultIconUrl(): string {
    return '';
  }
}

export class BuildConfigs extends Array<BuildConfig> {}

export function findBuildConfigByID(buildConfigs: BuildConfigs, params: Params): BuildConfig {
  const id = params['id'];
  if (id) {
    if (buildConfigs) {
      for (const buildConfig of buildConfigs) {
        if (id === buildConfig.name) {
          return buildConfig;
        }
      }
    }
  }
  return id;
}

export function combineBuildConfigAndBuilds(
  buildConfigs: BuildConfigs,
  builds: Builds,
): BuildConfigs {
  const map = {};
  const bcBuilds = {};
  if (builds) {
    builds.forEach((s) => {
      map[s.name] = s;
      const bcName = s.buildConfigName;
      if (bcName) {
        let list = bcBuilds[bcName];
        if (!list) {
          list = [];
          bcBuilds[bcName] = list;
        }
        list.push(s);
      }
    });
  }
  if (buildConfigs) {
    buildConfigs.forEach((bc) => {
      /*
        let lastVersionName = bc.lastBuildName;
        if (lastVersionName) {
          let build = map[lastVersionName];
          if (build) {
            bc.lastBuild = build;
          }
        }
*/
      const name = bc.name;
      if (name) {
        const list = bcBuilds[name];
        if (list) {
          bc.builds = list;
        }
      }
    });
  }
  return buildConfigs;
}

export function filterPipelines(buildConfigs: BuildConfigs): BuildConfigs {
  const answer = new BuildConfigs();
  buildConfigs.forEach((bc) => {
    if (bc.isPipeline) {
      answer.push(bc);
    }
  });
  return answer;
}

/**
 * returns a map of all the environments with the apps in each environment
 */
export function appInfos(buildConfigs: BuildConfigs): Map<string, EnvironmentApps> {
  const answer = new Map<string, EnvironmentApps>();
  buildConfigs.forEach((bc) => {
    const appEnv = bc.environmentApp;
    for (const environmentKey in appEnv) {
      if (environmentKey && appEnv[environmentKey]) {
        const app = appEnv[environmentKey];
        let env = answer[environmentKey];
        if (!env) {
          env = new EnvironmentApps();
          answer[environmentKey] = env;
        }
        if (!env.name) {
          env.name = app.environmentName;
        }
        env.apps[app.name] = app;
      }
    }
  });
  return answer;
}

/**
 * Keeps track of all the apps in each environment along with its name
 */
export class EnvironmentApps {
  apps: Map<string, AppInfo> = new Map<string, AppInfo>();

  name: string;
}
