import {
  AfterViewInit,
  Component,
  ElementRef,
  OnInit,
  ViewChild,
  ViewEncapsulation,
} from '@angular/core';
import { NgForm } from '@angular/forms';
import { Router } from '@angular/router';
import { Notification, Notifications, NotificationType } from 'ngx-base';
import { Context, Contexts } from 'ngx-fabric8-wit';
import { AuthenticationService, User, UserService } from 'ngx-login-client';
import { Subscription } from 'rxjs';
import {
  ExtProfile,
  GettingStartedService,
} from '../../getting-started/services/getting-started.service';
import { ProviderService } from '../../shared/account/provider.service';
import { GitHubService } from '../../space/create/codebases/services/github.service';
import { CopyService } from '../services/copy.service';
import { TenantService } from '../services/tenant.service';

@Component({
  encapsulation: ViewEncapsulation.None,
  selector: 'alm-update',
  templateUrl: 'tenant.component.html',
  styleUrls: ['./tenant.component.less'],
  providers: [CopyService, GettingStartedService, GitHubService, ProviderService, TenantService],
})
export class TenantComponent implements AfterViewInit, OnInit {
  @ViewChild('_templatesRepoBlob') templatesRepoBlobElement: HTMLElement;

  @ViewChild('_templatesRepo') templatesRepoElement: ElementRef;

  @ViewChild('_templatesRepoDir') templatesRepoDirElement: HTMLElement;

  @ViewChild('_boosterGitRef') boosterGitRefElement: HTMLElement;

  @ViewChild('_boosterGitRepo') boosterGitRepoElement: HTMLElement;

  @ViewChild('tenantForm') profileForm: NgForm;

  templatesRepoBlob: string;

  templatesRepo: string;

  templatesRepoDir: string;

  updateTenant: boolean = true;

  boosterGitRef: string;

  boosterGitRepo: string;

  context: Context;

  templatesRepoBlobInvalid: boolean = false;

  templatesRepoInvalid: boolean = false;

  templatesRepoDirInvalid: boolean = false;

  boosterGitRefInvalid: boolean = false;

  boosterGitRepoInvalid: boolean = false;

  loadedFormEmpty: boolean;

  gitHubLinked: boolean = false;

  imageUrl: string;

  loggedInUser: User;

  fullName: string;

  openShiftLinked: boolean = false;

  registrationCompleted: boolean = true;

  subscriptions: Subscription[] = [];

  token: string;

  username: string;

  url: string;

  private modifiedFromRequestParam: boolean = false;

  constructor(
    private auth: AuthenticationService,
    private gettingStartedService: GettingStartedService,
    private contexts: Contexts,
    private notifications: Notifications,
    private router: Router,
    private tenantService: TenantService,
    private userService: UserService,
  ) {
    this.subscriptions.push(contexts.current.subscribe((val) => (this.context = val)));

    if (userService.currentLoggedInUser.attributes) {
      this.loggedInUser = userService.currentLoggedInUser;
      this.setUserProperties(this.loggedInUser);
      this.subscriptions.push(
        auth.gitHubToken.subscribe((token) => {
          this.gitHubLinked = token !== undefined && token.length !== 0;
        }),
      );
      this.subscriptions.push(
        auth.isOpenShiftConnected(this.loggedInUser.attributes.cluster).subscribe((isConnected) => {
          this.openShiftLinked = isConnected;
        }),
      );
    }
  }

  ngAfterViewInit(): void {
    // Set focus
    if (this.getRequestParam('templatesRepoBlob') !== null) {
      this.setElementFocus(null, this.templatesRepoBlobElement);
    } else if (this.getRequestParam('boosterGitRef') !== null) {
      this.setElementFocus(null, this.boosterGitRefElement);
    }
  }

  ngOnInit() {
    this.token = this.auth.getToken();
  }

  ngOnDestroy(): void {
    this.subscriptions.forEach((sub) => {
      sub.unsubscribe();
    });
  }

  get isUpdateProfileDisabled(): boolean {
    if (this.modifiedFromRequestParam) {
      return false;
    }
    return (
      (!this.profileForm.dirty && !(!this.loadedFormEmpty && this.formValuesEmpty())) ||
      (this.templatesRepoBlobInvalid ||
        this.templatesRepoInvalid ||
        this.templatesRepoDirInvalid ||
        this.boosterGitRefInvalid ||
        this.boosterGitRepoInvalid)
    );
  }

  /**
   * Route to user profile
   */
  routeToProfile(): void {
    this.router.navigate(['/', this.context.user.attributes.username]);
  }

  /**
   * Set element focus to given HTML elment
   *
   * @param $event The triggered event
   * @param element The component or HTML element to set focus
   */
  setElementFocus($event: MouseEvent, element: any) {
    if (element instanceof HTMLElement) {
      (element as HTMLElement).focus();
    }
  }

  /**
   * Resets the profile back to empty values
   */

  resetProfile(): void {
    this.modifiedFromRequestParam = false;
    this.profileForm.reset();
    this.updateTenant = true;
  }

  /**
   * Update user profile
   */
  updateProfile(): void {
    const profile = this.getTransientProfile();
    let contextInformation = profile.contextInformation;
    if (!contextInformation) {
      contextInformation = {};
      profile.contextInformation = contextInformation;
    }

    let boosterCatalog = contextInformation['boosterCatalog'];
    if (!boosterCatalog) {
      boosterCatalog = {};
      contextInformation['boosterCatalog'] = boosterCatalog;
    }
    boosterCatalog['gitRef'] = this.boosterGitRef;
    boosterCatalog['gitRepo'] = this.boosterGitRepo;

    let tenantConfig = contextInformation['tenantConfig'];
    if (!tenantConfig) {
      tenantConfig = {};
      contextInformation['tenantConfig'] = tenantConfig;
    }
    tenantConfig['templatesRepoBlob'] = this.templatesRepoBlob;
    tenantConfig['templatesRepo'] = this.templatesRepo;
    tenantConfig['templatesRepoDir'] = this.templatesRepoDir;
    tenantConfig['updateTenant'] = this.updateTenant;

    this.subscriptions.push(
      this.gettingStartedService.update(profile).subscribe(
        (user) => {
          this.setUserProperties(user);

          this.subscriptions.push(
            this.tenantService.updateTenant().subscribe(
              (res) => {
                this.notifications.message({
                  message: `Tenant Updated!`,
                  type: NotificationType.SUCCESS,
                } as Notification);
                this.routeToProfile();
              },
              (error) => {
                this.handleError('Failed to update tenant', NotificationType.DANGER);
              },
            ),
          );
        },
        (error) => {
          if (error.status === 409) {
            this.handleError('Email already exists', NotificationType.DANGER);
          } else {
            this.handleError('Failed to update profile', NotificationType.DANGER);
          }
        },
      ),
    );
  }

  boosterGitRepoValidate(): void {
    this.boosterGitRefInvalid = !this.isUrlValid(this.boosterGitRepo);
  }

  templatesRepoValidate(): void {
    this.templatesRepoInvalid = !this.isUrlValid(this.templatesRepo);
  }

  // Private

  /**
   * lets default values from request parameters
   */
  private defaultValuesFromRequestParams() {
    this.modifiedFromRequestParam = false;
    this.templatesRepoBlob = this.defaultRequestParam('templatesRepoBlob', this.templatesRepoBlob);
    this.templatesRepo = this.defaultRequestParam('templatesRepo', this.templatesRepo);
    this.templatesRepoDir = this.defaultRequestParam('templatesRepoDir', this.templatesRepoDir);

    this.boosterGitRef = this.defaultRequestParam('boosterGitRef', this.boosterGitRef);
    this.boosterGitRepo = this.defaultRequestParam('boosterGitRepo', this.boosterGitRepo);
  }

  private defaultRequestParam(paramName: string, defaultValue: string): string {
    const answer = this.getRequestParam(paramName);
    if (answer) {
      this.modifiedFromRequestParam = true;
      return answer;
    }
    return defaultValue;
  }

  /**
   * Helper to retrieve request parameters
   *
   * @param name The request parameter to retrieve
   * @returns {any} The request parameter value or null
   */
  private getRequestParam(name: string): string {
    const param = new RegExp(`[?&]${encodeURIComponent(name)}=([^&]*)`).exec(
      window.location.search,
    );
    if (param != undefined) {
      return decodeURIComponent(param[1]);
    }
    return null;
  }

  /**
   * Get transient profile with updated properties
   *
   * @returns {ExtProfile} The updated transient profile
   */
  private getTransientProfile(): ExtProfile {
    const profile = this.gettingStartedService.createTransientProfile();
    delete profile.username;
    return profile;
  }

  /**
   * Set user properties
   *
   * @param user
   */
  private setUserProperties(user: User): void {
    this.modifiedFromRequestParam = false;
    if (user.attributes === undefined) {
      return;
    }

    let contextInformation = user.attributes['contextInformation'];
    if (!contextInformation) {
      contextInformation = {};
      user.attributes['contextInformation'] = contextInformation;
    }

    const boosterCatalog = contextInformation['boosterCatalog'];
    if (boosterCatalog) {
      this.boosterGitRef = boosterCatalog['gitRef'] || '';
      this.boosterGitRepo = boosterCatalog['gitRepo'] || '';
    }

    const tenantConfig = contextInformation['tenantConfig'];
    if (tenantConfig) {
      this.templatesRepoBlob = tenantConfig['templatesRepoBlob'] || '';
      this.templatesRepo = tenantConfig['templatesRepo'] || '';
      this.templatesRepoDir = tenantConfig['templatesRepoDir'] || '';
      this.updateTenant = tenantConfig['updateTenant'] || false;
    }

    this.loadedFormEmpty = this.formValuesEmpty();

    this.defaultValuesFromRequestParams();
  }

  /**
   * Helper to test if URL is valid
   *
   * @returns {boolean}
   */
  private isUrlValid(url: string): boolean {
    let trimmed = '';
    if (url) {
      trimmed = url.trim();
    }
    if (!trimmed) {
      return true;
    }
    return trimmed.startsWith('http:') || trimmed.startsWith('https:');
  }

  private handleError(error: string, type: NotificationType) {
    this.notifications.message({
      message: error,
      type,
    } as Notification);
  }

  private formValuesEmpty() {
    return !(
      this.boosterGitRef ||
      this.boosterGitRepo ||
      this.templatesRepoBlob ||
      this.templatesRepo ||
      this.templatesRepoDir ||
      !this.updateTenant
    );
  }
}
