import { ErrorHandler } from '@angular/core';
import { Space } from 'ngx-fabric8-wit';
import { Observable, ReplaySubject, Subject } from 'rxjs';
import { ExtProfile, ProfileService } from '../profile/profile.service';

export const RECENT_LENGTH: number = 8;

export interface RecentData<T> {
  recent: T[];
  isSaveRequired: boolean;
}

export abstract class RecentUtils<T> {
  protected _recent: Subject<T[]>;

  constructor(protected errorHandler: ErrorHandler, protected profileService: ProfileService) {
    this._recent = new ReplaySubject<T[]>(1);
  }

  get recent(): Observable<T[]> {
    return this._recent;
  }

  abstract compareElements<K = T>(t1: T, t2: K): boolean;

  onBroadcastChanged(
    changed: T,
    recent: T[],
    compareFn: ((a: T, b: T) => boolean) = this.compareElements,
  ): RecentData<T> {
    const index: number = recent.findIndex((t: T) => compareFn(t, changed));
    if (index === 0) {
      // continue only if changed is new, or requires a move within recent
      return { recent, isSaveRequired: false };
    }
    if (index > 0) {
      // if changed exists in recent, move it to the front
      recent.splice(index, 1);
      recent.unshift(changed);
    } else {
      // if changed is new to recent
      recent.unshift(changed);
      // trim recent if required to ensure it is length =< 8
      if (recent.length > RECENT_LENGTH) {
        recent.pop();
      }
    }
    return { recent, isSaveRequired: true };
  }

  onBroadcastDeleted<K>(
    deleted: K,
    recent: T[],
    compareFn: ((a: T, b: K) => boolean) = this.compareElements,
  ): RecentData<T> {
    const index: number = recent.findIndex((t: T) => compareFn(t, deleted));
    if (index === -1) {
      return { recent, isSaveRequired: false };
    }
    recent.splice(index, 1);
    return { recent, isSaveRequired: true };
  }

  saveProfile(patch: ExtProfile): void {
    this.profileService.silentSave(patch).subscribe(
      (): void => {},
      (err: string): void => {
        this.errorHandler.handleError(err);
      },
    );
  }
}
